"""Dataset module."""

from __future__ import unicode_literals

__all__ = ["from_list_of_dicts", "to_json", "to_jsonobject", "to_xml"]

from typing import Any, List, Optional, Set

import system.dataset
import system.date
from com.inductiveautomation.ignition.common import Dataset
from java.util import Date

from incendium.helper.types import AnyStr, DictStringAny


class _NanoXML(object):
    indent = None  # type: AnyStr
    root = None  # type: AnyStr

    def __init__(self, root="root", indent="\t"):
        # type: (AnyStr, AnyStr) -> None
        """Nano XML initializer.

        Args:
            root: The value of the XML root element.
            indent: Character(s) used for indentation.
        """
        super(_NanoXML, self).__init__()
        self.root = root
        self.indent = indent
        self._new_line = "\n"
        self._output = "<{root}>{new_line}".format(
            root=self.root, new_line=self._new_line
        )

    def add_element(self, name):
        # type: (AnyStr) -> None
        """Add an element to the XML document.

        Args:
            name: The name of the element.
        """
        self._output += "{indent}<{name}>{new_line}".format(
            indent=self.indent, name=name, new_line=self._new_line
        )

    def add_sub_element(self, name, value):
        # type: (AnyStr, AnyStr) -> None
        """Add a sub element to an element.

        Args:
            name: The name of the sub element.
            value: The value of the sub element.
        """
        self._output += "{indent}<{name}>{value}</{name}>{new_line}".format(
            indent=self.indent * 2,
            value=value,
            new_line=self._new_line,
            name=name,
        )

    def close_element(self, name):
        # type: (AnyStr) -> None
        """Close element.

        Args:
            name: The name of the element.
        """
        self._output += "{indent}</{name}>{new_line}".format(
            indent=self.indent, name=name, new_line=self._new_line
        )

    def to_string(self):
        # type: () -> AnyStr
        """Return the string representation of the XML document.

        Returns:
            The string representation of the XML document.
        """
        self._output += "</{}>".format(self.root)
        return self._output


def _format_object(obj):
    # type: (Any) -> Any
    """Format the object.

    Args:
        obj: The value to format.

    Returns:
        The representation of the object.
    """
    _obj = obj
    if isinstance(obj, Dataset):
        _obj = _to_jsonobject(obj)
    return _obj


def _format_value(obj, header=""):
    # type: (Any, AnyStr) -> AnyStr
    """Format the value to be properly represented in JSON.

    Args:
        obj: The value to format.
        header: Column name used for nested Datasets.

    Returns:
        The string representation of the value.
    """
    _obj = ""
    if obj is None:
        _obj = "null"
    elif isinstance(obj, basestring):
        _obj = '"{}"'.format(obj)
    elif isinstance(obj, Date):
        _obj = '"{}"'.format(system.date.format(obj, "yyyy-MM-dd'T'HH:mm:ss.SSSXXX"))
    elif isinstance(obj, Dataset):
        _obj = _to_json(obj, header, False)
    else:
        _obj = "{!r}".format(obj)
    return _obj


def _to_json(dataset, root=None, is_root=True):
    # type: (Dataset, Optional[AnyStr], bool) -> AnyStr
    """Return a string JSON representation of the Dataset.

    Private function.

    Args:
        dataset: The input dataset.
        root: The value of the header.
        is_root: True if we are at the root, False otherwise. Optional.

    Returns:
        The string JSON representation of the dataset.
    """
    headers = dataset.getColumnNames()
    columns = dataset.getColumnCount()
    rows = dataset.getRowCount()
    data = system.dataset.toPyDataSet(dataset)
    ret_str = ("{" if is_root and root is not None else "") + (
        '"{}":['.format(root) if root is not None else "["
    )
    col_count = 0

    for row_count, row in enumerate(data, start=1):
        ret_str += "{"
        for header in headers:
            col_count += 1
            val = _format_value(row[header], header)
            comma = "," if col_count < columns else ""
            if isinstance(row[header], Dataset):
                ret_str += "{}{}".format(val, comma)
            else:
                ret_str += '"{}":{}{}'.format(header, val, comma)
        ret_str += "{}{}".format("}", "," if row_count < rows else "")
        col_count = 0
    ret_str += "]"
    ret_str += "}" if is_root and root is not None else ""

    return ret_str


def _to_jsonobject(dataset):
    # type: (Dataset) -> List[DictStringAny]
    """Convert a Dataset into a Python list of dictionaries.

    Args:
        dataset: The input dataset.

    Returns:
        The Dataset as a Python object.
    """
    data = []
    headers = dataset.getColumnNames()
    row_count = dataset.getRowCount()

    for i in range(row_count):
        row_dict = {
            header: _format_object(dataset.getValueAt(i, header)) for header in headers
        }
        data.append(row_dict)

    return data


def from_list_of_dicts(list_of_dicts):
    # type: (List[DictStringAny]) -> Dataset
    """Safely convert a list of Python dictionaries into a Dataset.

    Args:
        list_of_dicts: The list of dictionaries.

    Returns:
        A Dataset representation of the list of dictionaries.
    """
    keys_set = set()  # type: Set[AnyStr]
    headers = list(keys_set.union(*(d.keys() for d in list_of_dicts)))
    data = []
    for dict_ in list_of_dicts:
        row = []
        for header in headers:
            if header in dict_.keys():
                row.append(dict_[header])
            else:
                row.append(None)
        data.append(row)
    return system.dataset.toDataSet(headers, data)


def to_json(dataset, root=None):
    # type: (Dataset, Optional[AnyStr]) -> AnyStr
    """Return a string JSON representation of the Dataset.

    Args:
        dataset: The input dataset.
        root: The value of the root. Optional.

    Returns:
        The string JSON representation of the dataset.
    """
    return _to_json(dataset, root)


def to_jsonobject(dataset):
    # type: (Dataset) -> List[DictStringAny]
    """Convert a Dataset into a Python list of dictionaries.

    Args:
        dataset: The input dataset.

    Returns:
        The Dataset as a Python object.
    """
    return _to_jsonobject(dataset)


def to_xml(dataset, root="root", element="row", indent="\t"):
    # type: (Dataset, AnyStr, AnyStr, AnyStr) -> AnyStr
    r"""Return a string XML representation of the Dataset.

    Args:
        dataset: The input dataset.
        root: The value of the root. If not provided, it defaults to
            "root". Optional.
        element: The value of the row. If not provided, it defaults to
            "row". Optional.
        indent: Current indentation. If not provided, it defaults to
            "\t". Optional.

    Returns:
        The string XML representation of the dataset.
    """
    headers = dataset.getColumnNames()
    row_count = dataset.getRowCount()
    xml = _NanoXML(root, indent)

    for i in range(row_count):
        xml.add_element(element)
        for header in headers:
            xml.add_sub_element(header, dataset.getValueAt(i, header))
        xml.close_element(element)

    return xml.to_string()



# coding=utf-8
# Copyright (C) 2020-2022
# Author: César Román
# Contact: cesar@thecesrom.dev
#
# Licensed under the MIT License.
#
# https://github.com/thecesrom/incendium/blob/HEAD/LICENSE
#

"""incendium.

incendium is a package that extends and wraps some functions from
Ignition Scripting API.

For more information, please refer to the Wiki.
https://github.com/thecesrom/incendium/wiki
"""



"""Types module."""

from typing import Any, Dict, Optional, Union

from java.lang import Exception as JavaException

AnyStr = Union[str, unicode]
DictIntStringAny = Dict[Union[int, str, unicode], Any]
DictStringAny = Dict[Union[str, unicode], Any]
InnerException = Optional[Union[Exception, JavaException]]
Number = Union[float, int, long]



"""Utility module."""

__all__ = [
    "get_function_name",
    "get_timer",
    "get_timestamp",
    "set_locale",
    "validate_form",
]

import traceback
from typing import Dict, Optional, Tuple, Union

import system.date
import system.util
from java.util import Date

from incendium import constants
from incendium.helper.types import AnyStr, Number
from incendium.user import IncendiumUser


def _format_error_message(counter, error_message, key):
    # type: (int, AnyStr, AnyStr) -> AnyStr
    """Format error message.

    Args:
        counter: Number of detected errors.
        error_message: Error message.
        key: Dictionary key.

    Returns:
        Formatted error message.
    """
    error_message += (
        constants.TABBED_LINE + key if counter == 1 else constants.NEW_TABBED_LINE + key
    )
    return error_message


def get_function_name():
    # type: () -> AnyStr
    """Get the name of the function last called.

    Returns:
        Function's name.
    """
    return traceback.extract_stack(None, 2)[0][2]


def get_timestamp(value):
    # type: (int) -> AnyStr
    """Get timestamp in "hh:mm:ss" format.

    Args:
        value: Time represented in seconds.

    Returns:
        Time elapsed represented by a string in the following format:
        "hh:mm:ss".
    """
    minutes, seconds = divmod(value, 60)
    hours, minutes = divmod(minutes, 60)
    return "{:02d}:{:02d}:{:02d}".format(hours, minutes, seconds)


def get_timer(date):
    # type: (Union[Date, long]) -> AnyStr
    """Get a timer with the time elapsed from value until now.

    Args:
        date: A date or a date represented in milliseconds.

    Returns:
         Time elapsed represented by a string in the following
         format: "hh:mm:ss".
    """
    date_1 = date if isinstance(date, Date) else system.date.fromMillis(date)
    date_2 = system.date.now()
    return get_timestamp(system.date.secondsBetween(date_1, date_2))


def set_locale(user):
    # type: (IncendiumUser) -> None
    """Set the Locale to the user's default Language.

    If none is configured, the default will be English (US).

    Args:
        user: IncendiumUser instance.
    """
    locale = (
        user.locale if user is not None and user.locale else constants.DEFAULT_LANGUAGE
    )

    system.util.setLocale(locale)


def validate_form(
    strings=None,  # type: Optional[Dict[AnyStr, AnyStr]]
    numbers=None,  # type: Optional[Dict[AnyStr, Number]]
    collections=None,  # type: Optional[Dict[AnyStr, Number]]
):
    # type: (...) -> Tuple[bool, AnyStr]
    """Perform a form validation.

    Args:
        strings: A dictionary containing all strings which must not be
            empty. Optional.
        numbers: A dictionary containing all numbers which must be
            greater than zero. Optional.
        collections: A dictionary containing all collections which must
            at least contain an element. Optional.

    Returns:
        A tuple (is_valid, error_message), where is_valid is True if all
        validation tests have passed, False otherwise, and error_message
        is the error message in case any validation test has failed.
    """
    is_valid = True
    error_message = constants.EMPTY_STRING
    counter = 0

    if strings:
        for key, str_val in strings.iteritems():
            if not str_val:
                counter += 1
                error_message = _format_error_message(counter, error_message, key)
                is_valid = False

    merged_dict = {}
    if numbers:
        merged_dict.update(numbers)
    if collections:
        merged_dict.update(collections)

    if merged_dict:
        for key, num_val in merged_dict.iteritems():
            if num_val is None or num_val <= 0:
                counter += 1
                error_message = _format_error_message(counter, error_message, key)
                is_valid = False

    return is_valid, error_message



"""User module."""

from __future__ import unicode_literals

__all__ = [
    "IncendiumUser",
    "get_emails",
    "get_user",
    "get_user_email_address",
    "get_user_first_name",
    "get_user_full_name",
    "get_users",
]

from typing import List, Optional

import system.security
import system.user
from com.inductiveautomation.ignition.common.user import ContactInfo, PyUser

from incendium.exceptions import ApplicationError
from incendium.helper.types import AnyStr


class IncendiumUser(object):
    """Wrapper class for Ignition's User object."""

    _roles = None  # type: List[AnyStr]
    _locale = None  # type: AnyStr
    _last_name = None  # type: AnyStr
    _first_name = None  # type: AnyStr
    _email = None  # type: Optional[AnyStr]
    _contact_info = None  # type: List[ContactInfo]

    def __init__(self, user):
        # type: (PyUser) -> None
        """User initializer.

        Args:
            user: Ignition's user object.
        """
        super(IncendiumUser, self).__init__()
        self._contact_info = user.getContactInfo()
        self._email = None
        self._first_name = user.get(user.FirstName)
        self._last_name = user.get(user.LastName)
        self._locale = user.getOrDefault(user.Language)
        self._roles = user.getRoles()

    @property
    def contact_info(self):
        # type: () -> List[ContactInfo]
        """Get User's ContactInfo.

        Returns:
            A sequence of ContactInfo objects.
        """
        return self._contact_info

    @property
    def email(self):
        # type: () -> List[AnyStr]
        """Get User's email address(es).

        Returns:
            User's email address(es).
        """
        return [ci.value for ci in self._contact_info if ci.contactType == "email"]

    @property
    def first_name(self):
        # type: () -> AnyStr
        """Get User's first name.

        Returns:
            User's first name.
        """
        return self._first_name

    @property
    def full_name(self):
        # type: () -> AnyStr
        """Get User's full name.

        Returns:
            User's full name.
        """
        return " ".join([self._first_name, self._last_name])

    @property
    def last_name(self):
        # type: () -> AnyStr
        """Get User's last name.

        Returns:
            User's last name.
        """
        return self._last_name

    @property
    def locale(self):
        # type: () -> AnyStr
        """Get User's preferred language.

        Returns:
            User's preferred language.
        """
        return self._locale

    @property
    def roles(self):
        # type: () -> List[AnyStr]
        """Get User's Roles.

        Returns:
            A list of Roles for this User.
        """
        return self._roles


def get_users(user_source="", filter_role=""):
    # type: (AnyStr, AnyStr) -> List[PyUser]
    """Get a list of PyUser objects from a User Source filtered by role.

    Args:
        user_source: The name of the User Source. If not provided, the
            default User Source will be consulted. Optional.
        filter_role: The name of the role. If provided, a list of PyUser
            objects for users that are assigned to a matching role will
            be retrieved, otherwise all users will be retrieved as
            PyUser objects. Optional.

    Returns:
        A list of PyUser objects.
    """
    users = system.user.getUsers(user_source)
    return (
        [user for user in users if filter_role in user.getRoles()]
        if filter_role
        else users
    )


def get_emails(user_source="", filter_role=""):
    # type: (AnyStr, AnyStr) -> List[AnyStr]
    """Get a list of email addresses from a User Source.

    Args:
        user_source: The name of the User Source. If not provided, the
            default User Source will be consulted. Optional.
        filter_role: The name of the role. If provided, a list of email
            addresses for users that are assigned to a matching role
            will be retrieved, otherwise all email addresses will be
            retrieved. Optional.

    Returns:
        A list of email addresses.
    """
    emails = set()
    users = [IncendiumUser(user) for user in get_users(user_source, filter_role)]
    for user in users:
        for email in user.email:
            emails.add(email)
    return sorted(list(emails))


def get_user(user_source="", failover=None):
    # type: (AnyStr, Optional[AnyStr]) -> IncendiumUser
    """Look up the logged-in User in a User Source.

    Args:
        user_source: The name of the User Source. If not provided, the
            default User Source will be consulted. Optional.
        failover: The name of the Failover Source. Optional.

    Returns:
        An IncendiumUser object.

    Raises:
        ApplicationError: If User is not found.
    """
    username = system.security.getUsername()

    user = system.user.getUser(user_source, username)

    if user is None and failover:
        user = system.user.getUser(failover, username)

    if user is None:
        raise ApplicationError("User was not found.")

    return IncendiumUser(user)


def get_user_email_address(user_source="", failover=None):
    # type: (AnyStr, Optional[AnyStr]) -> List[AnyStr]
    """Get the User's Email address(es).

    Args:
        user_source: The name of the User Source. If not provided, the
            default User Source will be consulted. Optional.
        failover: The name of the Fallback profile. Optional.

    Returns:
        The User's Email address(es).
    """
    return get_user(user_source, failover).email


def get_user_first_name(user_source="", failover=None):
    # type: (AnyStr, Optional[AnyStr]) -> AnyStr
    """Get the User's First Name.

    Args:
        user_source: The name of the User Source. If not provided, the
            default User Source will be consulted. Optional.
        failover: The name of the Fallback profile. Optional.

    Returns:
        The User's First Name.
    """
    return get_user(user_source, failover).first_name


def get_user_full_name(user_source="", failover=None):
    # type: (AnyStr, Optional[AnyStr]) -> AnyStr
    """Get the User's Full Name.

    Args:
        user_source: The name of the User Source. If not provided, the
            default User Source will be consulted. Optional.
        failover: The name of the Fallback profile. Optional.

    Returns:
        The User's Full Name.
    """
    return get_user(user_source, failover).full_name



"""GUI module."""

from __future__ import unicode_literals

__all__ = [
    "CURSOR_CROSSHAIR",
    "CURSOR_DEFAULT",
    "CURSOR_E_RESIZE",
    "CURSOR_HAND",
    "CURSOR_MOVE",
    "CURSOR_NE_RESIZE",
    "CURSOR_NW_RESIZE",
    "CURSOR_N_RESIZE",
    "CURSOR_SE_RESIZE",
    "CURSOR_SW_RESIZE",
    "CURSOR_S_RESIZE",
    "CURSOR_TEXT",
    "CURSOR_WAIT",
    "CURSOR_W_RESIZE",
    "authentication",
    "authorization",
    "confirm",
    "error",
    "info",
    "input",
    "warning",
]

from typing import Optional

import system.security
import system.util
from java.awt import GridLayout
from javax.swing import JLabel, JOptionPane, JPanel, JPasswordField, JTextField

from incendium import constants
from incendium.helper.types import AnyStr

# Cursor codes.
CURSOR_DEFAULT = 0
CURSOR_CROSSHAIR = 1
CURSOR_TEXT = 2
CURSOR_WAIT = 3
CURSOR_SW_RESIZE = 4
CURSOR_SE_RESIZE = 5
CURSOR_NW_RESIZE = 6
CURSOR_NE_RESIZE = 7
CURSOR_N_RESIZE = 8
CURSOR_S_RESIZE = 9
CURSOR_W_RESIZE = 10
CURSOR_E_RESIZE = 11
CURSOR_HAND = 12
CURSOR_MOVE = 13


def authentication(auth_profile="", title="Authentication"):
    # type: (AnyStr, AnyStr) -> bool
    """Open up a popup input dialog box.

    This dialog box will show a prompt message, and allow the user to
    type in their username and password. When the user is done, they can
    press "OK" or "Cancel". If "OK" is pressed, this function will
    attempt to validate the User credentials against the Authentication
    Profile. If "Cancel" is pressed, this function will return
    ``False``.

    Args:
        auth_profile: The name of the authentication profile to run
            against. Leaving this out will use the project's default
            profile. Optional.
        title: A title for the input box. This will be translated to the
            selected Locale. Optional.

    Returns:
        ``True`` if the user was validated, ``False`` otherwise.
    """
    options = [
        system.util.translate(constants.OK_TEXT),
        system.util.translate(constants.CANCEL_TEXT),
    ]

    panel = JPanel()

    labels = JPanel(GridLayout(0, 1, 2, 2))
    labels.add(JLabel("{}: ".format(system.util.translate("Username"))))
    labels.add(JLabel("{}: ".format(system.util.translate("Password"))))
    panel.add(labels)

    fields = JPanel(GridLayout(0, 1, 2, 2))
    username_field = JTextField(25)
    fields.add(username_field)
    password_field = JPasswordField(25)
    fields.add(password_field)
    panel.add(fields)

    choice = JOptionPane.showOptionDialog(
        None,
        panel,
        system.util.translate(title),
        JOptionPane.OK_CANCEL_OPTION,
        JOptionPane.PLAIN_MESSAGE,
        None,
        options,
        options[0],
    )

    valid = system.security.validateUser(
        username=username_field.getText(),
        password="".join(password_field.getPassword()),
        authProfile=auth_profile,
    )

    return choice == JOptionPane.OK_OPTION and valid


def authorization(role, auth_profile="", title="Athorization"):
    # type: (AnyStr, AnyStr, AnyStr) -> bool
    """Open up a popup input dialog box.

    This dialog box will show a prompt message, and allow the user to
    type in their username and password. When the user is done, they can
    press "OK" or "Cancel". If "OK" is pressed, this function will
    attempt to validate the User credentials against the Authentication
    Profile and verify if the User belongs to the ``role``. If "Cancel"
    is pressed, this function will return ``False``.

    Args:
        role: Required role.
        auth_profile: The name of the authentication profile to run
            against. Leaving this out will use the project's default
            profile. Optional.
        title: A title for the input box. This will be translated to the
            selected Locale. Optional.

    Returns:
        ``True`` if the user was validated, ``False`` otherwise.
    """
    has_role = False

    options = [
        system.util.translate(constants.OK_TEXT),
        system.util.translate(constants.CANCEL_TEXT),
    ]

    panel = JPanel()

    labels = JPanel(GridLayout(0, 1, 2, 2))
    labels.add(JLabel("{}: ".format(system.util.translate("Username"))))
    labels.add(JLabel("{}: ".format(system.util.translate("Password"))))
    panel.add(labels)

    fields = JPanel(GridLayout(0, 1, 2, 2))
    username_field = JTextField(25)
    fields.add(username_field)
    password_field = JPasswordField(25)
    fields.add(password_field)
    panel.add(fields)

    choice = JOptionPane.showOptionDialog(
        None,
        panel,
        system.util.translate(title),
        JOptionPane.OK_CANCEL_OPTION,
        JOptionPane.PLAIN_MESSAGE,
        None,
        options,
        options[0],
    )

    if choice == JOptionPane.OK_OPTION:
        user_roles = system.security.getUserRoles(
            username=username_field.getText(),
            password="".join(password_field.getPassword()),
            authProfile=auth_profile,
        )
        if user_roles is not None:
            has_role = role in user_roles

    return has_role


def confirm(message, title="Confirm", show_cancel=False):
    # type: (AnyStr, AnyStr, bool) -> Optional[bool]
    """Display a confirmation dialog box to the user.

    This will present the user with "Yes", "No" and "Cancel" options,
    and a custom message.

    Args:
        message: The message to display. This will be translated to the
            selected Locale.
        title: A title for the message box. This will be translated to
            the selected Locale. Optional.
        show_cancel: Show a cancel button in the dialog. Optional.

    Returns:
        True if the user selected "Yes", False if the user selected
        "No", None if the user selected "Cancel" or closes the dialog.
    """
    options = [
        system.util.translate(constants.YES_TEXT),
        system.util.translate(constants.NO_TEXT),
    ]

    if show_cancel:
        options.append(system.util.translate(constants.CANCEL_TEXT))

    choice = JOptionPane.showOptionDialog(
        None,
        system.util.translate(message),
        system.util.translate(title),
        JOptionPane.YES_NO_CANCEL_OPTION,
        JOptionPane.QUESTION_MESSAGE,
        None,
        options,
        options[0],
    )

    return (
        not bool(choice)
        if choice in [JOptionPane.YES_OPTION, JOptionPane.NO_OPTION]
        else None
    )


def error(message, title="Error", detail=None):
    # type: (AnyStr, AnyStr, Optional[AnyStr]) -> None
    """Display an error-style message box to the user.

    Args:
        message: The message to display in an error box. This will be
            translated to the selected Locale.
        title: A title for the error box. This will be translated to the
            selected Locale. Optional.
        detail: Additional text to display. This will be translated to
            the selected Locale. Optional.
    """
    if detail is None:
        msg = system.util.translate(message)
    else:
        msg = "\n".join([system.util.translate(message), system.util.translate(detail)])
    JOptionPane.showMessageDialog(
        None, msg, system.util.translate(title), JOptionPane.ERROR_MESSAGE
    )


def info(message, title="Information", detail=None):
    # type: (AnyStr, AnyStr, Optional[AnyStr]) -> None
    """Display an informational-style message popup box to the user.

    Args:
        message: The message to display. This will be translated to the
            selected Locale. Will accept html formatting.
        title: A title for the message box. This will be translated to
            the selected Locale. Optional.
        detail: Additional text to display. This will be translated to
            the selected Locale. Optional.
    """
    if detail is None:
        msg = system.util.translate(message)
    else:
        msg = "\n".join([system.util.translate(message), system.util.translate(detail)])
    JOptionPane.showMessageDialog(
        None,
        msg,
        system.util.translate(title),
        JOptionPane.INFORMATION_MESSAGE,
    )


def input(message, title="Input"):
    # type: (AnyStr, AnyStr) -> Optional[AnyStr]
    """Open up a popup input dialog box.

    This dialog box will show a prompt message, and allow the user to
    type in a string. When the user is done, they can press "OK" or
    "Cancel". If OK is pressed, this function will return with the value
    that they typed in. If Cancel is pressed, this function will return
    the value None.

    Args:
        message: The message to display. This will be translated to the
            selected Locale. Will accept html formatting.
        title: A title for the input box. This will be translated to the
            selected Locale. Optional.

    Returns:
        The string value that was entered in the input box.
    """
    options = [
        system.util.translate(constants.OK_TEXT),
        system.util.translate(constants.CANCEL_TEXT),
    ]

    panel = JPanel()
    label = JLabel("{}: ".format(system.util.translate(message)))
    panel.add(label)
    text_field = JTextField(25)
    panel.add(text_field)

    choice = JOptionPane.showOptionDialog(
        None,
        panel,
        system.util.translate(title),
        JOptionPane.OK_CANCEL_OPTION,
        JOptionPane.PLAIN_MESSAGE,
        None,
        options,
        options[0],
    )

    return text_field.getText() if choice == JOptionPane.OK_OPTION else None


def warning(message, title="Warning", detail=None):
    # type: (AnyStr, AnyStr, Optional[AnyStr]) -> None
    """Display a message to the user in a warning style popup dialog.

    Args:
        message: The message to display in a warning box. This will be
            translated to the selected Locale.
        title: A title for the warning box. This will be translated to
            the selected Locale. Optional.
        detail: Additional text to display. This will be translated to
            the selected Locale. Optional.
    """
    if detail is None:
        msg = system.util.translate(message)
    else:
        msg = "\n".join([system.util.translate(message), system.util.translate(detail)])
    JOptionPane.showMessageDialog(
        None, msg, system.util.translate(title), JOptionPane.WARNING_MESSAGE
    )



"""Date module."""

__all__ = ["compare", "get_first_day_of_month", "get_last_day_of_month"]

import system.date
from java.util import Date


def compare(date_1, date_2):
    # type: (Date, Date) -> int
    """Compare two dates.

    Args:
        date_1: The first date.
        date_2: The second date.

    Returns:
        0 if date_1 and date_2 are equal, -1 If date_2 is greater than
        date_1, 1 If date_1 is greater than date_2.
    """
    ret_val = 1

    if date_1 == date_2:
        ret_val = 0
    elif date_1 < date_2:
        ret_val = -1

    return ret_val


def get_first_day_of_month(date):
    # type: (Date) -> Date
    """Get first day of the month at midnight given the date.

    Args:
        date: The date.

    Returns:
        The date set to the first date of the month at midnight.
    """
    year = system.date.getYear(system.date.midnight(date))
    month = system.date.getMonth(date)
    return system.date.getDate(year, month, 1)


def get_last_day_of_month(date):
    # type: (Date) -> Date
    """Get last day of the month at 11:59:59 PM given the date.

    Args:
        date: The date.

    Returns:
        The date set to the last day of the month at 11:59:59 PM.
    """
    next_month = system.date.addMonths(system.date.midnight(date), 1)
    year = system.date.getYear(next_month)
    month = system.date.getMonth(next_month)
    return system.date.addSeconds(system.date.getDate(year, month, 1), -1)



"""Navigation module."""

from __future__ import unicode_literals

__all__ = ["swap_to", "swap_windows"]

from typing import List, Optional

import system.nav

from incendium.helper.types import AnyStr, DictStringAny


def _get_full_path(from_path, to_path):
    # type: (AnyStr, AnyStr) -> AnyStr
    """Return the full path of the window to swap to.

    The path will be relative to the path of the window to swap from.

    Args:
        from_path: The full path of the window to swap from.
        to_path: The full path or relative path of the window to swap
            to.

    Returns:
        The full path of the window to swap to.
    """
    current_directory = "."
    parent_directory = ".."
    path_separator = "/"
    _from = from_path.split(path_separator)
    _to = to_path.split(path_separator)
    path_parts = []
    if _to[0] not in (parent_directory, current_directory):
        path_parts = _to
    elif _to[0] == parent_directory:
        path_parts = _from[:-2] + _to[1:]
    elif _to[0] == current_directory:
        path_parts = _from[:-1] + _to[1:]

    full_path = []  # type: List[AnyStr]
    for path_part in path_parts:
        if path_part == parent_directory and full_path:
            full_path = full_path[:-1]
        elif path_part != current_directory:
            full_path.append(path_part)

    return path_separator.join(full_path)


def swap_windows(from_path, to_path, params=None):
    # type: (AnyStr, AnyStr, Optional[DictStringAny]) -> None
    """Perform a window swap.

    Args:
        from_path: The full path of the window to swap from.
        to_path: The full path or relative path of the window to swap
            to.
        params: A dictionary of parameters to pass into the window. The
            keys in the dictionary must match dynamic property names on
            the target window's root container. The values for each key
            will be used to set those properties. Optional.
    """
    _to_path = _get_full_path(from_path, to_path)
    if _to_path != from_path:
        system.nav.swapWindow(from_path, _to_path, params)


def swap_to(path, params=None):
    # type: (AnyStr, Optional[DictStringAny]) -> None
    """Perform a window swap.

    This will swap from the current main screen window to the window
    specified.

    Args:
        path: The full path or relative path of the window to swap to.
        params: A dictionary of parameters to pass into the window. The
            keys in the dictionary must match dynamic property names on
            the target window's root container. The values for each key
            will be used to set those properties. Optional.
    """
    swap_windows(system.nav.getCurrentWindow(), path, params)



"""Database module."""

__all__ = [
    "DisposableConnection",
    "InParam",
    "OutParam",
    "Param",
    "check",
    "execute_non_query",
    "get_data",
    "get_output_params",
    "get_return_value",
    "o_execute_non_query",
    "o_get_data",
]

from types import TracebackType
from typing import Any, List, Optional, Tuple, Type, TypedDict, Union

import system.db
import system.util
from com.inductiveautomation.ignition.common import BasicDataset
from java.lang import Thread

from incendium.helper.types import AnyStr, DictIntStringAny

_SProcResult = TypedDict(
    "_SProcResult",
    {
        "output_params": DictIntStringAny,
        "result_set": BasicDataset,
        "return_value": Optional[int],
        "update_count": int,
    },
)


class DisposableConnection(object):
    """Disposable Connection.

    A disposable connection enables a database connection in Ignition
    and disables it once the operation is completed to release
    resources.
    """

    def __init__(self, database, retries=3):
        # type: (AnyStr, int) -> None
        """Disposable Connection initializer.

        Args:
            database: The name of the database connection in
                Ignition.
            retries: The number of additional times to retry
                enabling the connection. Optional.
        """
        super(DisposableConnection, self).__init__()
        self._database = database
        self._retries = retries
        self._global_conn = "incendium_db_{}".format(database)

    @property
    def database(self):
        # type: () -> AnyStr
        """Get the name of the disposable connection."""
        return self._database

    @property
    def status(self):
        # type: () -> AnyStr
        """Get connection status."""
        connection_info = system.db.getConnectionInfo(self._database)
        return str(connection_info.getValueAt(0, "Status"))

    def __enter__(self):
        # type: () -> DisposableConnection
        """Enter the runtime context related to this object.

        Raises:
            IOError: If the connection's status reports as Faulted, or
                ir cannot be enabled.
        """
        system.db.setDatasourceEnabled(self._database, True)

        for _ in range(self._retries):
            Thread.sleep(1000)
            if self.status == "Valid":
                if self._global_conn not in system.util.globals:
                    system.util.globals[self._global_conn] = 0
                system.util.globals[self._global_conn] += 1
                break
            if self.status == "Faulted":
                raise IOError(
                    "The database connection {!r} is {}.".format(
                        self._database, self.status
                    )
                )
        else:
            raise IOError(
                "The database connection {!r} could not be enabled.".format(
                    self._database
                )
            )
        return self

    def __exit__(
        self,
        exc_type,  # type: Optional[Type[BaseException]]
        exc_val,  # type: Optional[BaseException]
        exc_tb,  # type: Optional[TracebackType]
    ):
        # type: (...) -> None
        """Exit the runtime context related to this object."""
        system.util.globals[self._global_conn] -= 1
        if system.util.globals[self._global_conn] == 0:
            system.db.setDatasourceEnabled(self._database, False)


class Param(object):
    """Base class used for defining [IN|OUT]PUT parameters."""

    def __init__(
        self,
        name_or_index,  # type: Union[int, AnyStr]
        type_code,  # type: int
        value=None,  # type: Optional[Any]
    ):
        # type: (...) -> None
        """Param object initializer.

        Args:
            name_or_index: Parameter name or index.
            type_code: Type code constant.
            value: Value of type type_code.
        """
        super(Param, self).__init__()
        self._name_or_index = name_or_index
        self._type_code = type_code
        self._value = value

    @property
    def name_or_index(self):
        # type: () -> Union[int, AnyStr]
        """Get value of name_or_index."""
        return self._name_or_index

    @property
    def type_code(self):
        # type: () -> int
        """Get value of type_code."""
        return self._type_code

    @property
    def value(self):
        # type: () -> Optional[Any]
        """Get value of value."""
        return self._value

    def __repr__(self):  # type: ignore[no-untyped-def]
        """Compute the "official" string representation."""
        return "{}(name_or_index={!r}, type_code={}, value={})".format(
            self.__class__.__name__,
            self.name_or_index,
            self.type_code,
            self.value,
        )

    def __str__(self):  # type: ignore[no-untyped-def]
        """Compute the "informal" string representation."""
        return "{!r}, {}, {}".format(self.name_or_index, self.type_code, self.value)


class InParam(Param):
    """Class used for declaring INPUT parameters."""

    def __init__(self, name_or_index, type_code, value):
        # type: (Union[int, AnyStr], int, Any) -> None
        """Create an instance of InParam.

        Args:
            name_or_index: Index (int starting at 1, not 0), or
                name (str).
            type_code: Type code constant from `system.db`.
            value: Value of type type_code.
        """
        super(InParam, self).__init__(name_or_index, type_code, value)


class OutParam(Param):
    """Class used for declaring OUTPUT parameters."""

    def __init__(self, name_or_index, type_code):
        # type: (Union[int, AnyStr], int) -> None
        """Create an instance of OutParam.

        Args:
            name_or_index: Index (int starting at 1, not 0), or name
                (str).
            type_code: Type code constant from `system.db`.
        """
        super(OutParam, self).__init__(name_or_index, type_code)


def _execute_sp(
    stored_procedure,  # type: AnyStr
    database="",  # type: AnyStr
    transaction=None,  # type: Optional[AnyStr]
    skip_audit=False,  # type: bool
    in_params=None,  # type: Optional[List[InParam]]
    out_params=None,  # type: Optional[List[OutParam]]
    get_out_params=False,  # type: bool
    get_result_set=False,  # type: bool
    get_ret_val=False,  # type: bool
    get_update_count=False,  # type: bool
    return_type_code=system.db.INTEGER,  # type: int
):
    # type: (...) -> _SProcResult
    """Execute a stored procedure against the connection.

    Args:
        stored_procedure: The name of the stored procedure to execute.
        database: The name of the database connection to execute
            against. If omitted or "", the project's default database
            connection will be used. Optional.
        transaction: A transaction identifier. If omitted, the call will
            be executed in its own transaction. Optional.
        skip_audit: A flag which, if set to True, will cause the
            procedure call to skip the audit system. Useful for some
            queries that have fields which won't fit into the audit log.
            Optional.
        in_params: A Dictionary containing INPUT parameters. Optional.
        out_params: A Dictionary containing OUTPUT parameters. Optional.
        get_out_params: A flag indicating whether to return OUTPUT
            parameters after execution. Optional.
        get_result_set: A flag indicating whether to return a dataset
            that is the resulting data of the stored procedure, if any.
            Optional.
        get_ret_val: A flag indicating whether to get the return value
            of the stored procedure Call. Optional.
        get_update_count: A flag indicating whether to return the number
            of rows modified by the stored procedure, or -1 if not
            applicable. Optional.
        return_type_code: The return value Type Code. Optional.

    Returns:
        Result dictionary.
    """
    call = system.db.createSProcCall(
        procedureName=stored_procedure,
        database=database,
        tx=transaction,
        skipAudit=skip_audit,
    )

    if in_params is not None:
        for i_param in in_params:
            call.registerInParam(
                i_param.name_or_index, i_param.type_code, i_param.value
            )

    if out_params is not None:
        for o_param in out_params:
            call.registerOutParam(o_param.name_or_index, o_param.type_code)

    if get_ret_val:
        call.registerReturnParam(return_type_code)

    system.db.execSProcCall(call)

    _out_params = {}
    if out_params is not None and get_out_params:
        for o_param in out_params:
            _out_params[o_param.name_or_index] = call.getOutParamValue(
                o_param.name_or_index
            )

    return {
        "output_params": _out_params,
        "result_set": call.getResultSet() if get_result_set else BasicDataset(),
        "return_value": call.getReturnValue() if get_ret_val else None,
        "update_count": call.getUpdateCount() if get_update_count else -1,
    }


def get_output_params(
    stored_procedure,  # type: AnyStr
    output,  # type: List[OutParam]
    database="",  # type: AnyStr
    transaction=None,  # type: Optional[AnyStr]
    params=None,  # type: Optional[List[InParam]]
):
    # type: (...) -> DictIntStringAny
    """Get the Output parameters from the Stored Procedure.

    Args:
        stored_procedure: The name of the stored procedure to execute.
        output: A list containing all OUTPUT parameters as OutParam
            objects.
        database: The name of the database connection to execute
            against. If omitted or "", the project's default database
            connection will be used. Optional.
        transaction: A transaction identifier. If omitted, the call will
            be executed in its own transaction. Optional.
        params: A list containing all INPUT parameters as InParam
            objects. Optional.

    Returns:
        A Python dictionary of OUTPUT parameters.
    """
    result = _execute_sp(
        stored_procedure,
        database=database,
        transaction=transaction,
        in_params=params,
        out_params=output,
        get_out_params=True,
    )

    return result["output_params"]


def check(stored_procedure, database="", params=None):
    # type: (AnyStr, AnyStr, Optional[List[InParam]]) -> Optional[bool]
    """Execute a stored procedure against the connection.

    This will return a flag set to TRUE or FALSE.

    Args:
        stored_procedure: The name of the stored procedure to execute.
        database: The name of the database connection to execute
            against. If omitted or "", the project's default database
            connection will be used. Optional.
        params: A Dictionary containing all INPUT parameters. Optional.

    Returns:
        The flag.
    """
    output = OutParam("flag", system.db.BIT)
    output_params = get_output_params(
        stored_procedure, output=[output], database=database, params=params
    )

    return output_params["flag"] if "flag" in output_params.iterkeys() else None


def execute_non_query(
    stored_procedure,  # type: AnyStr
    database="",  # type: AnyStr
    transaction=None,  # type: Optional[AnyStr]
    params=None,  # type: Optional[List[InParam]]
):
    # type: (...) -> int
    """Execute a stored procedure against the connection.

    Used for UPDATE, INSERT, and DELETE statements.

    Args:
        stored_procedure: The name of the stored procedure to execute.
        database: The name of the database connection to execute
            against. If omitted or "", the project's default database
            connection will be used. Optional.
        transaction: A transaction identifier. If omitted, the call will
            be executed in its own transaction. Optional.
        params: A list containing all INPUT parameters as InParam
            objects. Optional.

    Returns:
        The number of rows modified by the stored procedure, or -1 if
        not applicable.
    """
    result = _execute_sp(
        stored_procedure,
        database=database,
        transaction=transaction,
        in_params=params,
        get_update_count=True,
    )

    return result["update_count"]


def get_data(
    stored_procedure,  # type: AnyStr
    database="",  # type: AnyStr
    params=None,  # type: Optional[List[InParam]]
):
    # type: (...) -> BasicDataset
    """Get data by executing a stored procedure.

    Args:
        stored_procedure: The name of the stored procedure to execute.
        database: The name of the database connection to execute
            against. If omitted or "", the project's default database
            connection will be used. Optional.
        params: A list containing all INPUT parameters as InParam
            objects. Optional.

    Returns:
        A Dataset that is the resulting data of the stored procedure
        call.
    """
    result = _execute_sp(
        stored_procedure,
        database=database,
        in_params=params,
        get_result_set=True,
    )

    return result["result_set"]


def get_return_value(
    stored_procedure,  # type: AnyStr
    return_type_code,  # type: int
    database="",  # type: AnyStr
    transaction=None,  # type: Optional[AnyStr]
    params=None,  # type: Optional[List[InParam]]
):
    # type: (...) -> Optional[int]
    """Get the Return Value from the Stored Procedure.

    Args:
        stored_procedure: The name of the stored procedure to execute.
        return_type_code: The Type Code of the Return Value.
        database: The name of the database connection to execute
            against. If omitted or "", the project's default database
            connection will be used. Optional.
        transaction: A transaction identifier. If omitted, the call will
            be executed in its own transaction. Optional.
        params: A list containing all INPUT parameters as InParam
            objects. Optional.

    Returns:
        The return value.
    """
    result = _execute_sp(
        stored_procedure,
        database=database,
        transaction=transaction,
        in_params=params,
        get_ret_val=True,
        return_type_code=return_type_code,
    )

    return result["return_value"]


def o_execute_non_query(
    stored_procedure,  # type: AnyStr
    out_params,  # type: List[OutParam]
    database="",  # type: AnyStr
    transaction=None,  # type: Optional[AnyStr]
    in_params=None,  # type: Optional[List[InParam]]
):
    # type: (...) -> Tuple[int, DictIntStringAny]
    """Execute a stored procedure against the connection.

    Used for UPDATE, INSERT, and DELETE statements which return OUTPUT
    parameters.

    Args:
        stored_procedure: The name of the stored procedure to execute.
        out_params: A list containing all OUTPUT parameters as OutParam
            objects.
        database: The name of the database connection to execute
            against. If omitted or "", the project's default database
            connection will be used. Optional.
        transaction: A transaction identifier. If omitted, the call will
            be executed in its own transaction. Optional.
        in_params: A list containing all INPUT parameters as InParam
            objects. Optional.

    Returns:
        A tuple containing the number of rows modified by the stored
        procedure, or -1 if not applicable, and the OUTPUT parameters.
    """
    result = _execute_sp(
        stored_procedure,
        database=database,
        transaction=transaction,
        in_params=in_params,
        out_params=out_params,
        get_out_params=True,
        get_update_count=True,
    )

    return result["update_count"], result["output_params"]


def o_get_data(
    stored_procedure,  # type: AnyStr
    out_params,  # type: List[OutParam]
    database="",  # type: AnyStr
    in_params=None,  # type: Optional[List[InParam]]
):
    # type: (...) -> Tuple[BasicDataset, DictIntStringAny]
    """Get data by executing a stored procedure and OUTPUT parameters.

    Args:
        stored_procedure: The name of the stored procedure to execute.
        out_params: A list containing all OUTPUT parameters as OutParam
            objects.
        database: The name of the database connection to execute
            against. If omitted or "", the project's default database
            connection will be used. Optional.
        in_params: A list containing all INPUT parameters as InParam
            objects. Optional.

    Returns:
        A tuple containing a Dataset that is the resulting data of the
        stored procedure call, and the OUTPUT parameters.
    """
    result = _execute_sp(
        stored_procedure,
        database=database,
        in_params=in_params,
        out_params=out_params,
        get_out_params=True,
        get_result_set=True,
    )

    return result["result_set"], result["output_params"]



"""Tag module."""

__all__ = ["read", "write"]

from typing import Any

import system.tag
from com.inductiveautomation.ignition.common.model.values import BasicQualifiedValue

from incendium.helper.types import AnyStr


def read(tag_path):
    # type: (AnyStr) -> BasicQualifiedValue
    """Read the value of the Tags at the given paths.

    Note that this function will block until the read operation is
    complete or times out.

    Args:
        tag_path: Reads from the given tag path. If no property is
            specified in the path, the Value property is assumed.

    Returns:
        A qualified value. This object has three sub-members: value,
        quality, and timestamp.
    """
    values = system.tag.readBlocking([tag_path])
    return values[0]


def write(tag_path, value):
    # type: (AnyStr, Any) -> int
    """Write a value to a tag.

    Note that this function will block until the write operation is
    complete or times out.

    Args:
        tag_path: The path of the tag to write to.
        value: The value to write.

    Returns:
        0 if the operation failed immediately, 1 if it succeeded
        immediately, and 2 if it is pending.
    """
    quality_codes = system.tag.writeBlocking([tag_path], [value])
    return quality_codes[0].isGood()



"""Net module."""

from __future__ import unicode_literals

__all__ = [
    "send_high_priority_email",
    "send_html_email",
    "send_plain_text_email",
]

from typing import List

import system.net

from incendium import constants
from incendium.helper.types import AnyStr

HTML_ESCAPE_TABLE = {
    "&": "&amp;",
    '"': "&quot;",
    "'": "&apos;",
    ">": "&gt;",
    "<": "&lt;",
    "\n": "<br />",
    "\t": "&nbsp;" * 4,
}


def _html_escape(text):
    # type: (AnyStr) -> AnyStr
    """Escape an HTML file removing traces of offending characters.

    Args:
        text: The text to escape.

    Returns:
        The escaped text.
    """
    return "".join(HTML_ESCAPE_TABLE.get(c, c) for c in text)


def _send_email(subject, body, html, to, priority):
    # type: (AnyStr, AnyStr, bool, List[AnyStr], AnyStr) -> None
    """Send an email through the given SMTP server.

    Args:
        subject: The subject line for the email.
        body: The body text of the email.
        html: A flag indicating whether to send the email in HTML
            format.
        to: A list of email addresses to send to.
        priority: Priority for the message.
    """
    system.net.sendEmail(
        smtp=constants.SMTP,
        fromAddr=constants.SENDER,
        subject=subject,
        body=body,
        html=html,
        to=to,
        priority=priority,
    )


def send_html_email(subject, body, to, priority="3"):
    # type: (AnyStr, AnyStr, List[AnyStr], AnyStr) -> None
    """Send an email in HTML format.

    Args:
        subject: The subject line for the email.
        body: The body text of the email in HTML format.
        to: A list of email addresses to send to.
        priority: Priority of the message, from "1" to "5", with "1"
            being highest priority. Defaults to "3" (normal) priority.
            Optional.
    """
    _send_email(subject, body, True, to, priority)


def send_high_priority_email(subject, body, to):
    # type: (AnyStr, AnyStr, List[AnyStr]) -> None
    """Send a High Priority email.

    Args:
        subject: The subject line for the email.
        body: The body text of the email.
        to: A list of email addresses to send to.
    """
    send_html_email(subject, body, to, "1")


def report_error(subject, message, details, to):
    # type: (AnyStr, AnyStr, AnyStr, List[AnyStr]) -> None
    """Send an Error Report email message.

    Args:
        subject: The subject line for the email.
        message: The error message.
        details: The error details.
        to: A list of emails addresses to send to.
    """
    body = constants.ERROR_REPORT.format(message, _html_escape(details))
    send_high_priority_email(subject, body, to)


def send_plain_text_email(subject, body, to, priority="3"):
    # type: (AnyStr, AnyStr, List[AnyStr], AnyStr) -> None
    """Send an email in plain text format.

    Args:
        subject: The subject line for the email.
        body: The body text of the email in HTML format.
        to: A list of email addresses to send to.
        priority: Priority of the message, from "1" to "5", with "1"
            being highest priority. Defaults to "3" (normal) priority.
            Optional.
    """
    _send_email(subject, body, False, to, priority)



"""Exceptions module."""

from __future__ import unicode_literals

__all__ = ["ApplicationError", "TagError"]

from typing import Optional

from java.lang import Throwable

from incendium.helper.types import AnyStr, InnerException


class ApplicationError(Exception):
    """Application Error class."""

    cause = None  # type: Optional[Throwable]
    inner_exception = None  # type: InnerException
    message = None  # type: AnyStr

    def __init__(
        self,
        message,  # type: AnyStr
        inner_exception=None,  # type: InnerException
        cause=None,  # type: Optional[Throwable]
    ):
        # type: (...) -> None
        """Application Error initializer.

        Args:
            message: The error message.
            inner_exception: The inner Exception. Optional.
            cause: The cause of the Exception. Optional.
        """
        super(ApplicationError, self).__init__()
        self.message = message
        self.inner_exception = inner_exception
        self.cause = cause

    def __repr__(self):  # type: ignore[no-untyped-def]
        """Compute the "official" string representation."""
        return "{}(message={!r}, inner_exception={!r}, cause={!r})".format(
            self.__class__.__name__,
            self.message,
            self.inner_exception,
            self.cause,
        )

    def __str__(self):  # type: ignore[no-untyped-def]
        """Compute the "informal" string representation."""
        return "{!r}, {!r}, {!r}".format(self.message, self.inner_exception, self.cause)


class TagError(Exception):
    """Tag Error class."""

    message = None  # type: AnyStr

    def __init__(self, message):
        # type: (AnyStr) -> None
        """Tag Error initializer.

        Args:
            message: The error message.
        """
        super(TagError, self).__init__()
        self.message = message

    def __repr__(self):  # type: ignore[no-untyped-def]
        """Compute the "official" string representation."""
        return "{}(message={!r})".format(self.__class__.__name__, self.message)

    def __str__(self):  # type: ignore[no-untyped-def]
        """Compute the "informal" string representation."""
        return repr(self.message)
